from __future__ import unicode_literals, print_function, absolute_import, division, generators, nested_scopes

import datetime
import logging
import uuid

import os
from contextlib import contextmanager

import dateutil.parser
from sqlalchemy import Column, String, Boolean
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from commcare_export.writers import SqlMixin

logger = logging.getLogger(__name__)
repo_root = os.path.abspath(os.path.join(__file__, os.pardir, os.pardir))

Base = declarative_base()


class Checkpoint(Base):
    __tablename__ = 'commcare_export_runs'

    id = Column(String, primary_key=True)
    query_file_name = Column(String)
    query_file_md5 = Column(String)
    key = Column(String)
    project = Column(String)
    commcare = Column(String)
    since_param = Column(String)
    time_of_run = Column(String)
    final = Column(Boolean)

    def __repr__(self):
        return (
            "<Checkpoint("
            "id={r.id}, "
            "query_file_name={r.query_file_name}, "
            "query_file_md5={r.query_file_md5}, "
            "key={r.key}, "
            "project={r.project}, "
            "commcare={r.commcare}, "
            "since_param={r.since_param}, "
            "time_of_run={r.time_of_run}, "
            "final={r.final})>".format(r=self)
        )


@contextmanager
def session_scope(Session):
    """Provide a transactional scope around a series of operations."""
    session = Session()
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()


class CheckpointManager(SqlMixin):
    table_name = 'commcare_export_runs'
    migrations_repository = os.path.join(repo_root, 'migrations')

    def __init__(self, db_url, query, query_md5, project, commcare, key=None, poolclass=None):
        super(CheckpointManager, self).__init__(db_url, poolclass=poolclass)
        self.query = query
        self.query_md5 = query_md5
        self.project = project
        self.commcare = commcare
        self.key = key
        self.Session = sessionmaker(self.engine, expire_on_commit=False)

    def set_batch_checkpoint(self, checkpoint_time):
        self._set_checkpoint(checkpoint_time, False)

    def set_final_checkpoint(self):
        last_run = self.get_time_of_last_checkpoint(log_warnings=False)
        if last_run:
            self._set_checkpoint(dateutil.parser.parse(last_run), True)
            self._cleanup()

    def _set_checkpoint(self, checkpoint_time, final):
        logger.info('Setting %s checkpoint: %s', 'final' if final else 'batch', checkpoint_time)
        checkpoint_time = checkpoint_time or datetime.datetime.utcnow()
        with session_scope(self.Session) as session:
            session.add(Checkpoint(
                id=uuid.uuid4().hex,
                query_file_name=self.query,
                query_file_md5=self.query_md5,
                key=self.key,
                project=self.project,
                commcare=self.commcare,
                since_param=checkpoint_time.isoformat(),
                time_of_run=datetime.datetime.utcnow().isoformat(),
                final=final
            ))

    def create_checkpoint_table(self, revision='head'):
        from alembic import command, config
        cfg = config.Config(os.path.join(self.migrations_repository, 'alembic.ini'))
        cfg.set_main_option('script_location', self.migrations_repository)
        with self.engine.begin() as connection:
            cfg.attributes['connection'] = connection
            command.upgrade(cfg, revision)

    def _cleanup(self):
        with session_scope(self.Session) as session:
            session.query(Checkpoint).filter_by(
                final=False, query_file_md5=self.query_md5,
                project=self.project, commcare=self.commcare
            ).delete()

    def get_time_of_last_checkpoint(self, log_warnings=True):
        run = self.get_last_checkpoint()
        if run and log_warnings:
            self.log_warnings(run)
        return run.since_param if run else None

    def get_last_checkpoint(self):
        with session_scope(self.Session) as session:
            if self.key:
                run = self._get_last_checkpoint(
                    session, key=self.key,
                    project=self.project, commcare=self.commcare
                )
            else:
                run = self._get_last_checkpoint(
                    session, query_file_md5=self.query_md5,
                    project=self.project, commcare=self.commcare, key=self.key
                )
                if not run:
                    # Check for run without the args
                    run = self._get_last_checkpoint(session, query_file_md5=self.query_md5, key=self.key)
        return run

    def _get_last_checkpoint(self, session, **filters):
        return session.query(Checkpoint).filter_by(**filters)\
            .order_by(Checkpoint.time_of_run.desc()).first()

    def log_warnings(self, run):
        # type: (Checkpoint) -> None
        md5_mismatch = run.query_file_md5 != self.query_md5
        name_mismatch = run.query_file_name != self.query
        if md5_mismatch or name_mismatch:
            logger.warning(
                "Query differs from most recent checkpoint:\n"
                "From checkpoint:         name=%s, md5=%s\n"
                "From command line args:  name=%s, md5=%s\n",
                run.query_file_name, run.query_file_md5,
                self.query, self.query_md5
            )

    def list_checkpoints(self, limit=20):
        """List checkpoints filtered by:
        * file name
        * project
        * commcare
        * key

        Don't filter by MD5 on purpose.
        """
        with session_scope(self.Session) as session:
            query = session.query(Checkpoint)
            if self.query:
                query = query.filter(Checkpoint.query_file_name == self.query)
            if self.project:
                query = query.filter(Checkpoint.project == self.project)
            if self.commcare:
                query = query.filter(Checkpoint.commcare == self.commcare)
            if self.key:
                query = query.filter(Checkpoint.key == self.key)

            return query.order_by(Checkpoint.time_of_run.desc())[:limit]

    def update_checkpoint(self, run):
        with session_scope(self.Session) as session:
            session.merge(run)



from __future__ import unicode_literals, print_function, absolute_import, division, generators, nested_scopes
import re
from collections import defaultdict, namedtuple

from jsonpath_rw.lexer import JsonPathLexerError
from six.moves import xrange

from jsonpath_rw import jsonpath
from jsonpath_rw.parser import parse as parse_jsonpath

from commcare_export.exceptions import LongFieldsException
from commcare_export.map_format import compile_map_format_via
from commcare_export.minilinq import *

def take_while(pred, iterator):
    for v in iterator:
        if pred(v):
            yield v
        else:
            return

def drop_while(pred, iterator):
    for v in iterator:
        if not pred(v):
            yield v
            break

    for v in iterator:
        yield v

def without_empty_tail(cells):
    """
    Returns the prefix of a column that is not entirely empty.
    """
    return list(reversed(list(drop_while(lambda v: (not v) or (not v.value), reversed(cells)))))

def map_value(mappings_sheet, mapping_name, source_value):
    "From the mappings_sheet, replaces the source_value with appropriate output value"
    return source_value


def get_column_by_name(worksheet, column_name):
    # columns and rows are indexed from 1
    for col in xrange(1, worksheet.get_highest_column() + 1):
        value = worksheet.cell(row=1, column=col).value
        value = value.lower() if value else value
        if column_name == value:
            return without_empty_tail([
                worksheet.cell(row=i, column=col) for i in xrange(2, worksheet.get_highest_row() + 1)
            ])


def get_columns_by_prefix(worksheet, column_prefix):
    # columns and rows are indexed from 1
    for col in xrange(1, worksheet.get_highest_column() + 1):
        value = worksheet.cell(row=1, column=col).value
        if value and value.lower().startswith(column_prefix):
            yield value, without_empty_tail([
                worksheet.cell(row=i, column=col) for i in xrange(2, worksheet.get_highest_row() + 1)
            ])


def compile_mappings(worksheet):
    mapping_names = get_column_by_name(worksheet, "mapping name")
    sources       = extended_to_len(len(mapping_names), get_column_by_name(worksheet, "source"))
    destinations  = extended_to_len(len(mapping_names), get_column_by_name(worksheet, "destination"))

    mappings = defaultdict(lambda: defaultdict(lambda: None))
    
    for mapping_name, source, dest in zip(mapping_names, sources, destinations):
        if mapping_name and source:
            mappings[mapping_name.value][source.value] = dest.value if dest else None

    return mappings

def compile_filters(worksheet, mappings=None):
    filter_names  = [cell.value for cell in get_column_by_name(worksheet, 'filter name') or []]

    if not filter_names:
        return []

    filter_values = extended_to_len(len(filter_names), [cell.value for cell in get_column_by_name(worksheet, 'filter value') or []])
    return zip(filter_names, filter_values)

def extended_to_len(desired_len, some_list, value=None):
    return [some_list[i] if i < len(some_list) else value
            for i in xrange(0, desired_len)]


def _get_safe_source_field(source_field):
    def _safe_node(node):
        try:
            parse_jsonpath(node)
        except Exception:
            # quote nodes with special characters
            return '"{}"'.format(node)
        else:
            return node

    try:
        parse_jsonpath(source_field)
    except Exception:
        source_field = '.'.join([
            _safe_node(node) if node else node
            for node in source_field.split('.')
        ])
        if source_field.endswith('.'):
            raise Exception("Blank node path: {}".format(source_field))

    return Reference(source_field)


def compile_field(field, source_field, alternate_source_fields=None, map_via=None, format_via=None, mappings=None):
    expr = _get_safe_source_field(source_field)

    if alternate_source_fields:
        expr = Apply(Reference('or'), expr, *[Reference(alt_field) for alt_field in alternate_source_fields])
    if map_via:
        expr = compile_map_format_via(expr, map_via)

    if format_via:
        expr = compile_map_format_via(expr, format_via)

    if mappings and field in mappings:
        expr = compile_mapped_field(mappings[field], expr)

    return expr


def compile_mapped_field(field_mappings, field_expression):
    # quote the ref in case it has special chars
    quoted_field = Apply(Reference('join'), Literal(''), Literal('"'), field_expression, Literal('"'))
    # produce the mapping reference i.e. 'mapping."X"'
    mapping_ref = Apply(Reference('join'), Literal('.'), Literal('mapping'), quoted_field)
    # apply the reference to the field mappings to get the final value
    mapped_value = FlatMap(source=Literal([field_mappings]), body=Reference(mapping_ref), name='mapping')
    return Apply(Reference('default'), mapped_value, field_expression)


def _get_alternate_source_fields_from_csv(worksheet, num_fields):
    def _clean_csv_field(field):
        if field and field.value:
            return [val.strip() for val in field.value.split(',')]

    alt_source_col = get_column_by_name(worksheet, 'alternate source fields')

    if alt_source_col:
        alt_source_fields = extended_to_len(num_fields, alt_source_col)
        return [_clean_csv_field(field) for field in alt_source_fields]


def _get_alternate_source_fields_from_columns(worksheet, num_fields):
    matching_columns = sorted(get_columns_by_prefix(worksheet, 'alternate source field'), key=lambda x: x[0])
    alt_source_cols = [
        extended_to_len(num_fields, [cell.value if cell else cell for cell in alt_col])
        for col_name, alt_col in matching_columns
    ]
    # transpose columns to rows
    alt_srouce_fields = map(list, zip(*alt_source_cols))
    return [list(filter(None, fields)) for fields in alt_srouce_fields]


def get_alternate_source_fields(worksheet, num_fields):
    return (
        _get_alternate_source_fields_from_csv(worksheet, num_fields)
        or _get_alternate_source_fields_from_columns(worksheet, num_fields)
        or extended_to_len(num_fields, [])
    )


def compile_fields(worksheet, mappings=None):
    fields = without_empty_tail(get_column_by_name(worksheet, 'field') or [])

    if not fields:
        return []

    source_fields = extended_to_len(len(fields), get_column_by_name(worksheet, 'source field') or [])
    map_vias      = extended_to_len(len(fields), get_column_by_name(worksheet, 'map via') or [])
    format_vias   = extended_to_len(len(fields), get_column_by_name(worksheet, 'format via') or [])

    alternate_source_fields = get_alternate_source_fields(worksheet, len(fields))

    args = zip(fields, source_fields, alternate_source_fields, map_vias, format_vias)
    return [
        compile_field(
            field=field.value,
            source_field=source_field.value,
            alternate_source_fields=alt_source_fields,
            map_via=map_via.value if map_via else None,
            format_via=format_via.value if format_via else None,
            mappings=mappings
        )
        for field, source_field, alt_source_fields, map_via, format_via in args
    ]

def split_leftmost(jsonpath_expr):
    if isinstance(jsonpath_expr, jsonpath.Child):
        further_leftmost, rest = split_leftmost(jsonpath_expr.left)
        return further_leftmost, rest.child(jsonpath_expr.right)
    elif isinstance(jsonpath_expr, jsonpath.Descendants):
        further_leftmost, rest = split_leftmost(jsonpath_expr.left)
        return further_leftmost, jsonpath.Descendants(rest, jsonpath_expr.right)
    else:
        return (jsonpath_expr, jsonpath.This())


def compile_source(worksheet):
    """
    Compiles just the part of the Excel Spreadsheet that
    indicates the API endpoint to hit along with optional filters
    and an optional JSONPath within that endpoint, 

    For example, this spreadsheet
    
    Data Source                    Filter Name   Filter Value        Include Referenced Items
    -----------------------------  ------------  ------------------  --------------------------
    form[*].form.child_questions   app_id        <app id>            cases
                                   xmlns.exact   <some form xmlns>

    Should fetch from api/form?app_id=<app id>&xmlns.exact=<some form xmlns>&cases__full=true
    and then iterate (FlatMap) over all child questions.

    :return: tuple of the 'data source' expression and the 'root doc expression'.
        'data source': The MiniLinq that calls 'api_data' function to get data from CommCare
        'root doc expression': The MiniLinq that is applied to each doc, can be None.
    """

    data_source_column = get_column_by_name(worksheet, 'data source')
    if not data_source_column:
        raise Exception('Sheet has no "Data Source" column.')
    data_source_str = data_source_column[0].value
    filters = compile_filters(worksheet)
    include_referenced_items = [cell.value for cell in (get_column_by_name(worksheet, 'include referenced items') or [])]

    data_source, data_source_jsonpath = split_leftmost(parse_jsonpath(data_source_str))
    maybe_redundant_slice, remaining_jsonpath = split_leftmost(data_source_jsonpath)

    # The leftmost _must_ be of type Fields with one field and will pull out the first field
    if not isinstance(data_source, jsonpath.Fields) or len(data_source.fields) > 1:
        raise Exception('Bad value for data source: %s' % str(data_source))

    data_source = data_source.fields[0]

    if isinstance(maybe_redundant_slice, jsonpath.Slice):
        data_source_jsonpath = remaining_jsonpath

    api_query_args = [Reference("api_data"), Literal(data_source)]
    
    if not filters:
        if include_referenced_items:
            api_query_args.append(Literal(None)) # Pad the argument list if we have further args; keeps tests and user code more readable at the expense of this conditional
    else:
        api_query_args.append(Literal(dict(filters)))

    if include_referenced_items:
        api_query_args.append(Literal(include_referenced_items))

    api_query = Apply(*api_query_args)

    if data_source_jsonpath is None or isinstance(data_source_jsonpath, jsonpath.This) or isinstance(data_source_jsonpath, jsonpath.Root):
        return api_query, None
    else:
        return api_query, Reference(str(data_source_jsonpath))


def parse_sheet(worksheet, mappings=None):
    mappings = mappings or {}
    source_expr, root_doc_expr = compile_source(worksheet)

    table_name_column = get_column_by_name(worksheet, 'table name')
    if table_name_column:
        output_table_name = table_name_column[0].value
    else:
        output_table_name = worksheet.title
    output_headings = get_column_by_name(worksheet, 'field')
    output_fields = compile_fields(worksheet, mappings=mappings)

    if not output_fields:
        headings = []
        source = source_expr
        body = None
    else:
        headings = [Literal(output_heading.value) for output_heading in output_headings]
        source = source_expr
        body = List(output_fields)

    return SheetParts(
        output_table_name,
        headings,
        source,
        body,
        root_doc_expr,
    )


class SheetParts(namedtuple('SheetParts', 'name headings source body root_expr')):
    def __new__(cls, name, headings, source, body, root_expr=None):
        return super(SheetParts, cls).__new__(cls, name, headings, source, body, root_expr)


def parse_workbook(workbook):
    """
    Returns a MiniLinq corresponding to the Excel configuration, which
    consists of the following sheets:

    1. "Mappings" a sheet with three columns that defines simple lookup table functions
       A. MappingName - the name by which this mapping is referenced
       B. Source - the value to match
       C. Destination - the value to return

    2. Each other sheet represents one data table to emit
    """
    mappings_sheet = workbook.get_sheet_by_name('Mappings')
    mappings = compile_mappings(mappings_sheet) if mappings_sheet else None

    emit_sheets = [sheet_name for sheet_name in workbook.get_sheet_names() if sheet_name != 'Mappings']

    parsed_sheets = []
    for sheet in emit_sheets:
        try:
            sheet_parts = parse_sheet(workbook.get_sheet_by_name(sheet), mappings)
        except Exception as e:
            logger.warning('Ignoring sheet "{}": {}'.format(sheet, str(e)))
            continue

        parsed_sheets.append(sheet_parts)

    return parsed_sheets


def compile_queries(parsed_sheets, missing_value, combine_emits):
    # group sheets by source
    sheets_by_source = []
    for sheet in parsed_sheets:
        # Not easy to implement hashing on MiniLinq objects so can't use a dict
        for source, sheets in sheets_by_source:
            if sheet.source == source:
                sheets.append(sheet)
                break
        else:
            sheets_by_source.append((sheet.source, [sheet]))

    queries = []
    for source, sheets in sheets_by_source:
        if len(sheets) > 1:
            if combine_emits:
                queries.append(get_multi_emit_query(source, sheets, missing_value))
            else:
                queries.extend([
                    get_single_emit_query(sheet, missing_value)
                    for sheet in sheets
                ])
        else:
            queries.append(get_single_emit_query(sheets[0], missing_value))
    return queries


def get_multi_emit_query(source, sheets, missing_value):
    """Multiple `Emit` expressions using the same data source.
    For this we reverse the `Map` so that we apply each `Emit`
    repeatedly for each doc produced by the data source.
    """
    emits = []
    multi_query = Filter(  # the filter here is to prevent accumulating a `[None]` value for each doc
        predicate=Apply(
            Reference("filter_empty"),
            Reference("$")
        ),
        source=Map(
            source=source,
            body=List(emits)
        )
    )

    for sheet in sheets:
        # if there is no root expression then we just reference the whole document with `this`
        root_expr = sheet.root_expr or Reference("`this`")
        emits.append(
            Emit(
                table=sheet.name,
                headings=sheet.headings,
                source=Map(
                    source=root_expr,
                    body=sheet.body
                ),
                missing_value=missing_value
            )
        )

    return multi_query


def get_single_emit_query(sheet, missing_value):
    """Single `Emit` for the data source to we can just
    apply the `Emit` once with the source expression being
    the data source.
    """
    def _get_source(source, root_expr):
        if root_expr:
            return FlatMap(
                source=source,
                body=root_expr
            )
        else:
            return source

    return Emit(
        table=sheet.name,
        headings=sheet.headings,
        source=Map(
            source=_get_source(sheet.source, sheet.root_expr),
            body=sheet.body
        ),
        missing_value=missing_value
    )


def check_field_length(parsed_sheets, max_column_length):
    long_fields = defaultdict(list)
    for sheet in parsed_sheets:
        for header in sheet.headings:
            if len(header.v) > max_column_length:
                long_fields[sheet.name].append(header.v)

    if long_fields:
        raise LongFieldsException(long_fields, max_column_length)


def get_queries_from_excel(workbook, missing_value=None, combine_emits=False, max_column_length=None):
    parsed_sheets = parse_workbook(workbook)
    if max_column_length:
        check_field_length(parsed_sheets, max_column_length)
    queries = compile_queries(parsed_sheets, missing_value, combine_emits)
    return List(queries) if len(queries) > 1 else queries[0]



from __future__ import unicode_literals, print_function, absolute_import, division, generators, nested_scopes

import logging
from collections import OrderedDict

import backoff
import requests
from requests.auth import AuthBase
from requests.auth import HTTPDigestAuth

AUTH_MODE_PASSWORD = 'password'
AUTH_MODE_APIKEY = 'apikey'

try:
    from urllib.request import urlopen
    from urllib.parse import urlparse, urlencode, parse_qs
except ImportError:
    from urlparse import urlparse, parse_qs
    from urllib import urlopen, urlencode

import commcare_export
from commcare_export.repeatable_iterator import RepeatableIterator

logger = logging.getLogger(__file__)

LATEST_KNOWN_VERSION='0.5'


def on_backoff(details):
    _log_backoff(details, 'Waiting for retry.')


def on_giveup(details):
    _log_backoff(details, 'Giving up.')


def _log_backoff(details, action_message):
    details['__suffix'] = action_message
    logger.warn("Request failed after {tries} attempts ({elapsed:.1f}s). {__suffix}".format(**details))


def is_client_error(ex):
    if hasattr(ex, 'response') and ex.response:
        return 400 <= ex.response.status_code < 500
    return False


class CommCareHqClient(object):
    """
    A connection to CommCareHQ for a particular version, project, and user.
    """

    def __init__(self, url, project, username, password,
                 auth_mode=AUTH_MODE_PASSWORD, version=LATEST_KNOWN_VERSION, checkpoint_manager=None):
        self.version = version
        self.url = url
        self.project = project
        self._checkpoint_manager = checkpoint_manager
        self.__auth = self._get_auth(username, password, auth_mode)
        self.__session = None

    def _get_auth(self, username, password, mode):
        if mode == AUTH_MODE_PASSWORD:
            return HTTPDigestAuth(username, password)
        elif mode == AUTH_MODE_APIKEY:
            return ApiKeyAuth(username, password)
        else:
            raise Exception('Unknown auth mode: %s' % mode)

    @property
    def session(self):
        if self.__session == None:
            self.__session = requests.Session()
            self.__session.headers.update({
                'User-Agent': 'commcare-export/%s' % commcare_export.__version__
            })
        return self.__session

    @session.setter
    def session(self, session):
        """Used for overriding the session in unit tests"""
        self.__session = session

    @property
    def api_url(self):
        return '%s/a/%s/api/v%s' % (self.url, self.project, self.version)

    @backoff.on_exception(
        backoff.expo, requests.exceptions.RequestException,
        max_time=300, giveup=is_client_error,
        on_backoff=on_backoff, on_giveup=on_giveup
    )
    def get(self, resource, params=None):
        """
        Gets the named resource.

        Currently a bit of a vulnerable stub that works
        for this particular use case in the hands of a trusted user; would likely
        want this to work like (or via) slumber.
        """
        logger.debug("Fetching batch: %s", params)
        resource_url = '%s/%s/' % (self.api_url, resource)
        response = self.session.get(resource_url, params=params, auth=self.__auth, timeout=60)
        response.raise_for_status()
        return response.json()
            
    def iterate(self, resource, paginator, params=None):
        """
        Assumes the endpoint is a list endpoint, and iterates over it
        making a lot of assumptions that it is like a tastypie endpoint.
        """
        params = dict(params or {})
        def iterate_resource(resource=resource, params=params):
            more_to_fetch = True
            last_batch_ids = set()

            while more_to_fetch:
                batch = self.get(resource, params)
                total_count = int(batch['meta']['total_count']) if batch['meta']['total_count'] else 'unknown'
                logger.debug('Received %s-%s of %s', 
                             batch['meta']['offset'], 
                             int(batch['meta']['offset'])+int(batch['meta']['limit']),
                             total_count)
                
                if not batch['objects']:
                    more_to_fetch = False
                else:
                    for obj in batch['objects']:
                        if obj['id'] not in last_batch_ids:
                            yield obj

                    if batch['meta']['next']:
                        last_batch_ids = {obj['id'] for obj in batch['objects']}
                        params = paginator.next_page_params_from_batch(batch)
                        if not params:
                            more_to_fetch = False
                    else:
                        more_to_fetch = False

                self.checkpoint(paginator, batch)
                
        return RepeatableIterator(iterate_resource)

    def checkpoint(self, paginator, batch):
        from commcare_export.commcare_minilinq import DatePaginator
        if self._checkpoint_manager and isinstance(paginator, DatePaginator):
            since_date = paginator.get_since_date(batch)
            self._checkpoint_manager.set_batch_checkpoint(checkpoint_time=since_date)


class MockCommCareHqClient(object):
    """
    An in-memory mock of the hq client, instantiated
    with a simple mapping of resource and params to results.

    Since dictionaries are not hashable, the mapping is
    written as a pair of tuples, handled appropriately
    internallly.

    MockCommCareHqClient({
        'forms': [
            (
                {'_search': 'test1'},
                [
                   ... objects ...
                ]
            ),
        ]
    })
    """    
    def __init__(self, mock_data):
        self.mock_data = dict([(resource, dict([(urlencode(OrderedDict(sorted(params.items()))), result) for params, result in resource_results]))
                              for resource, resource_results in mock_data.items()])

    def iterate(self, resource, paginator, params=None):
        logger.debug('Mock client call to resource "%s" with params "%s"', resource, params)
        return self.mock_data[resource][urlencode(OrderedDict(sorted(params.items())))]


class ApiKeyAuth(AuthBase):
    def __init__(self, username, apikey):
        self.username = username
        self.apikey = apikey

    def __eq__(self, other):
        return all([
            self.username == getattr(other, 'username', None),
            self.apikey == getattr(other, 'apikey', None)
        ])

    def __hash__(self):
        return hash((self.username, self.apikey))

    def __ne__(self, other):
        return not self == other

    def __call__(self, r):
        r.headers['Authorization'] = 'apikey %s:%s' % (self.username, self.apikey)
        return r




class LongFieldsException(Exception):
    def __init__(self, long_fields, max_length):
        self.long_fields = long_fields
        self.max_length = max_length

    @property
    def message(self):
        message = ''
        for table, headers in self.long_fields.items():
            message += (
                'Table "{}" has field names longer than the maximum allowed for this database ({}):\n'.format(
                table, self.max_length
            ))
            for header in headers:
                message += '    {}\n'.format(header)

        message += '\nPlease adjust field names to be within the maximum length limit of {}'.format(self.max_length)
        return message



from __future__ import unicode_literals, print_function, absolute_import, division, generators, nested_scopes

import argparse
import getpass
import io
import json
import logging
import os.path
import sys

import dateutil.parser
import requests
from six.moves import input

from commcare_export import excel_query
from commcare_export import writers
from commcare_export.utils import get_checkpoint_manager
from commcare_export.commcare_hq_client import CommCareHqClient, LATEST_KNOWN_VERSION
from commcare_export.commcare_minilinq import CommCareHqEnv
from commcare_export.env import BuiltInEnv, JsonPathEnv, EmitterEnv
from commcare_export.exceptions import LongFieldsException
from commcare_export.minilinq import MiniLinq
from commcare_export.repeatable_iterator import RepeatableIterator
from commcare_export.version import __version__

EXIT_STATUS_ERROR = 1

logger = logging.getLogger(__name__)

commcare_hq_aliases = {
    'local': 'http://localhost:8000',
    'prod': 'https://www.commcarehq.org'
}


class Argument(object):
    def __init__(self, name, *args, **kwargs):
        self.name = name.replace('-', '_')
        self._args = ['--{}'.format(name)] + list(args)
        self._kwargs = kwargs

    @property
    def default(self):
        return self._kwargs.get('default')

    def add_to_parser(self, parser, **additional_kwargs):
        additional_kwargs.update(self._kwargs)
        parser.add_argument(*self._args, **additional_kwargs)


CLI_ARGS = [
        Argument('version', default=False, action='store_true',
                 help='Print the current version of the commcare-export tool.'),
        Argument('query', help='JSON or Excel query file. If omitted, JSON string is read from stdin.'),
        Argument('dump-query', default=False, action='store_true'),
        Argument('commcare-hq', default='prod',
                 help='Base url for the CommCare HQ instance e.g. https://www.commcarehq.org'),
        Argument('api-version', default=LATEST_KNOWN_VERSION),
        Argument('project'),
        Argument('username'),
        Argument('password', help='Enter password, or if using apikey auth-mode, enter the api key.'),
        Argument('auth-mode', default='password', choices=['password', 'apikey'],
                 help='Use "digest" auth, or "apikey" auth (for two factor enabled domains).'),
        Argument('since', help='Export all data after this date. Format YYYY-MM-DD or YYYY-MM-DDTHH:mm:SS'),
        Argument('until', help='Export all data up until this date. Format YYYY-MM-DD or YYYY-MM-DDTHH:mm:SS'),
        Argument('start-over', default=False, action='store_true',
                 help='When saving to a SQL database; the default is to pick up since the last success. This disables that.'),
        Argument('profile'),
        Argument('verbose', default=False, action='store_true'),
        Argument('output-format', default='json', choices=['json', 'csv', 'xls', 'xlsx', 'sql', 'markdown'],
                 help='Output format'),
        Argument('output', metavar='PATH', default='reports.zip', help='Path to output; defaults to `reports.zip`.'),
        Argument('strict-types', default=False, action='store_true',
                 help="When saving to a SQL database don't allow changing column types once they are created."),
        Argument('missing-value', default=None, help="Value to use when a field is missing from the form / case."),
        Argument('batch-size', default=1000, help="Number of records to process per batch."),
        Argument('checkpoint-key', help="Use this key for all checkpoints instead of the query file MD5 hash "
                                        "in order to prevent table rebuilds after a query file has been edited."),
    ]


def main(argv):
    parser = argparse.ArgumentParser('commcare-export', 'Output a customized export of CommCareHQ data.')
    for arg in CLI_ARGS:
        arg.add_to_parser(parser)

    try:
        args = parser.parse_args(argv)
    except UnicodeDecodeError:
        for arg in argv:
            try:
                arg.encode('utf-8')
            except UnicodeDecodeError:
                sys.stderr.write(u"ERROR: Argument '%s' contains unicode characters. "
                                 u"Only ASCII characters are supported.\n" % unicode(arg, 'utf-8'))
        sys.exit(1)

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG, 
                            format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
    else:
        logging.basicConfig(level=logging.WARN,
                            format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')

    logging.getLogger('alembic').setLevel(logging.WARN)
    logging.getLogger('backoff').setLevel(logging.FATAL)

    if args.version:
        print('commcare-export version {}'.format(__version__))
        exit(0)

    if not args.project:
        print('commcare-export: error: argument --project is required')
        exit(1)

    if args.profile:
        # hotshot is gone in Python 3
        import hotshot
        import hotshot.stats
        profile = hotshot.Profile(args.profile)
        profile.start()

    try:
        exit(main_with_args(args))
    finally:
        if args.profile:
            profile.close()
            stats = hotshot.stats.load(args.profile)
            stats.strip_dirs()
            stats.sort_stats('cumulative', 'calls')
            stats.print_stats(100)


def _get_query(args, writer):
    # Reads as excel if it is a file name that looks like excel, otherwise reads as JSON,
    # falling back to parsing arg directly as JSON, and finally parsing stdin as JSON
    if args.query:
        return _get_query_from_file(
            args.query,
            args.missing_value,
            writer.supports_multi_table_write,
            writer.max_column_length,
        )
    else:
        try:
            return MiniLinq.from_jvalue(json.loads(sys.stdin.read()))
        except Exception as e:
            raise Exception(
                "Failure reading query from console input. "
                "Try using the '--query' parameter to pass your query as an Excel file", e
            )

def _get_query_from_file(query_arg, missing_value, combine_emits, max_column_length):
    if os.path.exists(query_arg):
        if os.path.splitext(query_arg)[1] in ['.xls', '.xlsx']:
            import openpyxl
            workbook = openpyxl.load_workbook(query_arg)
            return excel_query.get_queries_from_excel(workbook, missing_value, combine_emits, max_column_length)
        else:
            with io.open(query_arg, encoding='utf-8') as fh:
                return MiniLinq.from_jvalue(json.loads(fh.read()))


def _get_writer(output_format, output, strict_types):
    if output_format == 'xlsx':
        return writers.Excel2007TableWriter(output)
    elif output_format == 'xls':
        return writers.Excel2003TableWriter(output)
    elif output_format == 'csv':
        if not output.endswith(".zip"):
            print("WARNING: csv output is a zip file, but "
                  "will be written to %s" % output)
            print("Consider appending .zip to the file name to avoid confusion.")
        return writers.CsvTableWriter(output)
    elif output_format == 'json':
        return writers.JValueTableWriter()
    elif output_format == 'markdown':
        return writers.StreamingMarkdownTableWriter(sys.stdout)
    elif output_format == 'sql':
        # Output should be a connection URL
        # Writer had bizarre issues so we use a full connection instead of passing in a URL or engine
        return writers.SqlTableWriter(output, strict_types)
    else:
        raise Exception("Unknown output format: {}".format(output_format))


def get_date_params(args, checkpoint_manager):
    if args.start_over and checkpoint_manager:
        logger.warn('Ignoring all checkpoints and re-fetching all data from CommCare.')

    if not args.since and not args.start_over and checkpoint_manager:
        args.since = checkpoint_manager.get_time_of_last_checkpoint()

        if args.since:
            logger.debug('Last successful checkpoint was %s', args.since)
        else:
            logger.warn('No successful runs found, and --since not specified: will import ALL data')

    since = dateutil.parser.parse(args.since) if args.since else None
    until = dateutil.parser.parse(args.until) if args.until else None
    return since, until


def _get_api_client(args, checkpoint_manager, commcarehq_base_url):
    return CommCareHqClient(
        url=commcarehq_base_url,
        project=args.project,
        username=args.username,
        password=args.password,
        auth_mode=args.auth_mode,
        version=args.api_version,
        checkpoint_manager=checkpoint_manager
    )


def _get_checkpoint_manager(args):
    if not os.path.exists(args.query):
        logger.warning("Checkpointing disabled for non file-based query")
    elif args.since or args.until:
        logger.warning("Checkpointing disabled when using '--since' or '--until'")
    else:
        checkpoint_manager = get_checkpoint_manager(args)
        checkpoint_manager.create_checkpoint_table()
        return checkpoint_manager


def main_with_args(args):
    writer = _get_writer(args.output_format, args.output, args.strict_types)

    try:
        query = _get_query(args, writer)
    except LongFieldsException as e:
        print(e.message)
        return EXIT_STATUS_ERROR

    if not query:
        print('Query file not found: %s' % args.query)
        return EXIT_STATUS_ERROR

    if args.dump_query:
        print(json.dumps(query.to_jvalue(), indent=4))
        return

    checkpoint_manager = None
    if writer.support_checkpoints:
        checkpoint_manager = _get_checkpoint_manager(args)

    if not args.username:
        args.username = input('Please provide a username: ')

    if not args.password:
        # Windows getpass does not accept unicode
        args.password = getpass.getpass()

    commcarehq_base_url = commcare_hq_aliases.get(args.commcare_hq, args.commcare_hq)
    api_client = _get_api_client(args, checkpoint_manager, commcarehq_base_url)

    since, until = get_date_params(args, checkpoint_manager)
    if since:
        logger.debug('Starting from %s', args.since)
    env = (
            BuiltInEnv({'commcarehq_base_url': commcarehq_base_url})
            | CommCareHqEnv(api_client, since=since, until=until, page_size=args.batch_size)
            | JsonPathEnv({})
            | EmitterEnv(writer)
    )

    with env:
        try:
            lazy_result = query.eval(env)
            if lazy_result is not None:
                # evaluate lazy results
                for r in lazy_result:
                    list(r) if r else r
        except requests.exceptions.RequestException as e:
            if e.response.status_code == 401:
                print("\nAuthentication failed. Please check your credentials.")
                return
            else:
                raise


    if checkpoint_manager:
        checkpoint_manager.set_final_checkpoint()

    if args.output_format == 'json':
        print(json.dumps(list(writer.tables.values()), indent=4, default=RepeatableIterator.to_jvalue))


def entry_point():
    main(sys.argv[1:])


if __name__ == '__main__':
    entry_point()
